import os

from pcs import settings

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall
from pcs_test.tools.command_env.mock_runner import CheckStdinEqualXml
from pcs_test.tools.fixture_cib import modify_cib
from pcs_test.tools.fixture_crm_mon import complete_state
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str

# pylint: disable=too-many-arguments

DEFAULT_WAIT_TIMEOUT = 10
WAIT_TIMEOUT_EXPIRED_RETURNCODE = 124
AGENT_FILENAME_MAP = {
    "ocf:heartbeat:Dummy": "resource_agent_ocf_heartbeat_dummy.xml",
    "ocf:pacemaker:Dummy": "resource_agent_ocf_pacemaker_dummy.xml",
    "ocf:pacemaker:remote": "resource_agent_ocf_pacemaker_remote.xml",
    "ocf:heartbeat:IPaddr2": "resource_agent_ocf_heartbeat_ipaddr2.xml",
    "ocf:pacemaker:booth-site": "resource_agent_ocf_pacemaker_booth-site.xml",
    "ocf:pacemaker:Stateful": "resource_agent_ocf_pacemaker_stateful_ocf_1.1.xml",
    "systemd:chronyd": "resource_agent_systemd_chronyd.xml",
    "stonith:fence_unfencing": "stonith_agent_fence_unfencing.xml",
}

RULE_IN_EFFECT_RETURNCODE = 0
RULE_EXPIRED_RETURNCODE = 110
RULE_NOT_YET_IN_EFFECT_RETURNCODE = 111


class PcmkShortcuts:
    # pylint: disable=too-many-public-methods
    def __init__(self, calls):
        self.__calls = calls
        self.default_wait_timeout = DEFAULT_WAIT_TIMEOUT
        self.default_wait_error_returncode = WAIT_TIMEOUT_EXPIRED_RETURNCODE

    def can_fence_history_manage(
        self,
        name="runner.pcmk.can_fence_history_manage",
        stderr="--history --cleanup --broadcast",
        instead=None,
    ):
        """
        Create a call to check if fence_history is supported by stonith_admin

        string name -- key of the call
        string stderr -- stonith_admin help text
        string instead -- key of call instead of which this new call is to be
            placed
        """
        self.__calls.place(
            name,
            RunnerCall(["stonith_admin", "--help-all"], stderr=stderr),
            instead=instead,
        )

    def can_fence_history_status(
        self,
        name="runner.pcmk.can_fence_history_status",
        stderr="--fence-history",
        instead=None,
        env=None,
    ):
        """
        Create a call to check if fence_history is supported by crm_mon

        string name -- key of the call
        string stderr -- crm_mon help text
        string instead -- key of call instead of which this new call is to be
            placed
        dict env -- CommandRunner environment variables
        """
        self.__calls.place(
            name,
            RunnerCall(["crm_mon", "--help-all"], stderr=stderr, env=env),
            instead=instead,
        )

    def fence_history_get(
        self,
        name="runner.pcmk.fence_history_get",
        node=None,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create call for getting plain text fencing history.

        string name -- key of the call
        string node -- a node to get a history from
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["stonith_admin", "--history", node, "--verbose"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def fence_history_cleanup(
        self,
        name="runner.pcmk.fence_history_cleanup",
        node=None,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create call for cleaning fencing history up.

        string name -- key of the call
        string node -- a node to clean a history from
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["stonith_admin", "--history", node, "--cleanup"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def fence_history_update(
        self,
        name="runner.pcmk.fence_history_update",
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create call for updating fencing history.

        string name -- key of the call
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["stonith_admin", "--history", "*", "--broadcast"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def load_state(
        self,
        name="runner.pcmk.load_state",
        filename="crm_mon.minimal.xml",
        resources=None,
        nodes=None,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create call for loading pacemaker state.

        string name -- key of the call
        string filename -- points to file with the status in the content
        string resources -- xml - resources section, will be put to state
        string nodes -- xml - nodes section, will be put to state
        string stdout -- crm_mon's stdout
        string stderr -- crm_mon's stderr
        int returncode -- crm_mon's returncode
        dict env -- CommandRunner environment variables
        """
        if (resources or nodes) and (stdout or stderr or returncode):
            raise AssertionError(
                "Cannot specify resources or nodes when stdout, stderr or "
                "returncode is specified"
            )

        command = ["crm_mon", "--one-shot", "--inactive", "--output-as", "xml"]

        if stdout or stderr or returncode:
            self.__calls.place(
                name,
                RunnerCall(
                    command,
                    stdout=stdout,
                    stderr=stderr,
                    returncode=returncode,
                    env=env,
                ),
            )
            return

        with open(rc(filename)) as a_file:
            state_xml = a_file.read()

        self.__calls.place(
            name,
            RunnerCall(
                command,
                stdout=etree_to_str(
                    complete_state(state_xml, resources, nodes)
                ),
                env=env,
            ),
        )

    def load_state_plaintext(
        self,
        name="runner.pcmk.load_state_plaintext",
        inactive=True,
        verbose=False,
        fence_history=False,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create a call for loading plaintext pacemaker status

        str name -- key of the call
        bool inactive -- pass --inactive flag to crm_mon
        bool verbose -- pass flags for increased verbosity to crm_mon
        bool fence_history -- pass the flag for getting fence history to crm_mon
        str stdout -- crm_mon's stdout
        str stderr -- crm_mon's stderr
        int returncode -- crm_mon's returncode
        dict env -- CommandRunner environment variables
        """
        flags = ["--one-shot"]
        if inactive:
            flags.append("--inactive")
        if verbose:
            flags.extend(
                ["--show-detail", "--show-node-attributes", "--failcounts"]
            )
            if fence_history:
                flags.append("--fence-history=3")
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_mon"] + flags,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def load_ticket_state_plaintext(
        self,
        name="runner.pcmk.load_ticket_state_plaintext",
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create a call for loading plaintext tickets status

        str name -- key of the call
        str stdout -- crm_ticket's stdout
        str stderr -- crm_ticket's stderr
        int returncode -- crm_ticket's returncode
        dict env -- CommandRunner environment variables
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_ticket", "--details"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def list_agents_standards(
        self,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
        name="runner.pcmk.list_agents_standards",
    ):
        """
        Create a call for listing agents standards
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_resource", "--list-standards"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def list_agents_ocf_providers(
        self,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
        name="runner.pcmk.list_agents_ocf_providers",
    ):
        """
        Create a call for listing agents ocf providers
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_resource", "--list-ocf-providers"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def list_agents_for_standard_and_provider(
        self,
        standard_provider,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
        name="runner.pcmk.list_agents_for_standard_and_provider",
    ):
        """
        Create a call for listing agents of given standard and provider
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_resource", "--list-agents", standard_provider],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def load_agent(
        self,
        name="runner.pcmk.load_agent",
        agent_name="ocf:heartbeat:Dummy",
        agent_filename=None,
        agent_is_missing=False,
        stdout=None,
        stderr=None,
        instead=None,
        env=None,
    ):
        """
        Create call for loading resource agent metadata.

        string name -- key of the call
        string agent_name
        string agent_filename -- points to file with the agent metadata in the
            content
        bool agent_is_missing -- create a response as if the agent was missing
        string instead -- key of call instead of which this new call is to be
            placed
        dict env -- CommandRunner environment variables
        """
        if env:
            env = dict(env)
        else:
            env = {}
        env["PATH"] = ":".join(
            [
                settings.fence_agent_binaries,
                "/bin",
                "/usr/bin",
            ]
        )

        if agent_filename:
            agent_metadata_filename = agent_filename
        elif agent_name in AGENT_FILENAME_MAP:
            agent_metadata_filename = AGENT_FILENAME_MAP[agent_name]
        elif not stdout and not agent_is_missing:
            raise AssertionError(
                (
                    "Filename with metadata of agent '{0}' not specified.\n"
                    "Please specify file with metadata for agent:\n"
                    "  a) explicitly for this test:"
                    " config.runner.pcmk.load_agent(agent_name='{0}',"
                    " filename='FILENAME_HERE.xml')\n"
                    "  b) implicitly for agent '{0}' in 'AGENT_FILENAME_MAP' in"
                    " '{1}'\n"
                    "Place agent metadata into '{2}FILENAME_HERE.xml'\n"
                    "Or define metadata directly in 'stdout' argument."
                ).format(agent_name, os.path.realpath(__file__), rc(""))
            )

        if agent_is_missing:
            if stderr is None:
                stderr = (
                    f"Agent {agent_name} not found or does not support "
                    "meta-data: Invalid argument (22)\n"
                    f"Metadata query for {agent_name} failed: Input/output "
                    "error\n"
                )
            self.__calls.place(
                name,
                RunnerCall(
                    ["crm_resource", "--show-metadata", agent_name],
                    stdout="",
                    stderr=stderr,
                    returncode=74,
                    env=env,
                ),
                instead=instead,
            )
            return

        if not stdout:
            with open(rc(agent_metadata_filename)) as a_file:
                stdout = a_file.read()
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_resource", "--show-metadata", agent_name],
                stdout=stdout,
                stderr=stderr,
                env=env,
            ),
            instead=instead,
        )

    def load_fake_agent_metadata(
        self,
        name="runner.pcmk.load_fake_agent_metadata",
        agent_name="pacemaker-fenced",
        stdout=None,
        stderr="",
        returncode=0,
        instead=None,
        before=None,
    ):
        """
        Create a call for loading fake agent metadata - usually metadata
        provided by pacemaker daemon

        string name -- the key of this call
        string agent_name -- name of the fake agent
        string stdout -- fake agent stdout, default metadata if None
        string stderr -- fake agent stderr
        int returncode -- fake agent returncode
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        """
        name_to_metadata_file = {
            "pacemaker-based": "based_metadata.xml",
            "pacemaker-controld": "controld_metadata.xml",
            "pacemaker-fenced": "fenced_metadata.xml",
            "pacemaker-schedulerd": "schedulerd_metadata.xml",
        }
        if stdout is None:
            with open(rc(name_to_metadata_file[agent_name])) as a_file:
                stdout = a_file.read()
        agent_path = settings.__dict__[agent_name.replace("-", "_")]
        self.__calls.place(
            name,
            RunnerCall(
                [agent_path, "metadata"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def local_node_name(
        self,
        name="runner.pcmk.local_node_name",
        instead=None,
        before=None,
        node_name="",
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create a call for crm_node --name

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string node_name -- resulting node name
        string stdout -- crm_node's stdout
        string stderr -- crm_node's stderr
        int returncode -- crm_node's returncode
        """
        if node_name and (stdout or stderr or returncode):
            raise AssertionError(
                "Cannot specify node_name when stdout, stderr or returncode is "
                "specified"
            )
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_node", "--name"],
                stdout=(node_name if node_name else stdout),
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def resource_cleanup(
        self,
        name="runner.pcmk.cleanup",
        instead=None,
        before=None,
        resource=None,
        node=None,
        strict=False,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create a call for crm_resource --cleanup

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be cleaned
        string node -- the name of the node where resources should be cleaned
        bool strict -- strict mode of 'crm_resource cleanup' enabled?
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        cmd = ["crm_resource", "--cleanup"]
        if resource:
            cmd.extend(["--resource", resource])
        if node:
            cmd.extend(["--node", node])
        if strict:
            cmd.extend(["--force"])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def resource_refresh(
        self,
        name="runner.pcmk.refresh",
        instead=None,
        before=None,
        resource=None,
        node=None,
        strict=False,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create a call for crm_resource --refresh

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be cleaned
        string node -- the name of the node where resources should be cleaned
        bool strict -- strict mode of 'crm_resource refresh' enabled?
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        cmd = ["crm_resource", "--refresh"]
        if resource:
            cmd.extend(["--resource", resource])
        if node:
            cmd.extend(["--node", node])
        if strict:
            cmd.extend(["--force"])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def resource_move(
        self,
        name="runner.pcmk.resource_move",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create a call for crm_resource --move

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be moved
        string node -- the name of a destination node
        bool master -- limit move to master role
        string lifetime -- lifetime of the created moving constraint
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        dict env -- CommandRunner environment variables
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--move"
        self._resource_move_ban_clear(**all_args)

    def resource_ban(
        self,
        name="runner.pcmk.resource_ban",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create a call for crm_resource --ban

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be banned
        string node -- the name of a destination node
        bool master -- limit ban to master role
        string lifetime -- lifetime of the created banning constraint
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--ban"
        self._resource_move_ban_clear(**all_args)

    def resource_clear(
        self,
        name="runner.pcmk.resource_clear",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        expired=None,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create a call for crm_resource --clear

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be unmoved/unbanned
        string node -- the name of a destination node
        bool master -- limit clearing to master role
        bool epired -- clear only expired moves and bans
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        dict env -- CommandRunner environment variables
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--clear"
        self._resource_move_ban_clear(**all_args)

    def _resource_move_ban_clear(
        self,
        name,
        action,
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        expired=None,
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        cmd = ["crm_resource", action]
        if resource:
            cmd.extend(["--resource", resource])
        if node:
            cmd.extend(["--node", node])
        if master:
            cmd.extend(["--master"])
        if lifetime:
            cmd.extend(["--lifetime", lifetime])
        if expired:
            cmd.extend(["--expired"])
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
            before=before,
            instead=instead,
        )

    def can_clear_expired(
        self, name="runner.pcmk.can_clear_expired", stderr="--expired"
    ):
        """
        Create call which check --expired is supported by crm_resource

        string name -- key of the call
        string stderr -- crm_resource help text
        """
        self.__calls.place(
            name,
            RunnerCall(["crm_resource", "--help-all"], stderr=stderr),
        )

    def wait(
        self, name="runner.pcmk.wait", stderr="", returncode=None, timeout=None
    ):
        """
        Create call for waiting to pacemaker idle

        string name -- key of the call
        string stderr -- stderr of wait command
        int returncode -- returncode of the wait command, defaults to 0 if
            stderr is empty and to 124 if stderr is not empty
        """
        if returncode is None:
            returncode = self.default_wait_error_returncode if stderr else 0

        cmd = ["crm_resource", "--wait"]
        if timeout != 0:
            cmd.append(
                "--timeout={0}".format(
                    timeout
                    if timeout is not None
                    else self.default_wait_timeout
                )
            )
        self.__calls.place(
            name,
            RunnerCall(cmd, stderr=stderr, returncode=returncode),
        )

    def is_resource_digests_supported(
        self,
        name="runner.pcmk.is_resource_digests_supported",
        is_supported=True,
    ):
        """
        Create call for `crm_resource --help-all`. If support_digest is True, option
        --digest is included in command output.

        name -- key of the call
        is_supported -- flags which decides if digests are supported
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_resource", "--help-all"],
                stdout="--digests" if is_supported else "",
                stderr="",
                returncode=0,
            ),
        )

    def resource_digests(
        self,
        resource_id,
        node_name,
        name="runner.pcmk.resource_digests",
        args=None,
        stdout="",
        stderr="",
        returncode=0,
    ):
        """
        Create call for crm_resource digests

        resource_id -- id of a resource
        node_name -- name of the node where the resource is running
        name -- key of the call
        args -- additional arguments for crm_resource
        stdout -- crm_resource's stdout
        stderr -- crm_resource's stderr
        returncode -- crm_resource's returncode
        """
        if args is None:
            args = []

        self.__calls.place(
            name,
            RunnerCall(
                [
                    "crm_resource",
                    "--digests",
                    "--resource",
                    resource_id,
                    "--node",
                    node_name,
                    "--output-as",
                    "xml",
                    *args,
                ],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def verify(
        self,
        name="runner.pcmk.verify",
        cib_tempfile=None,
        stderr=None,
        verbose=False,
        env=None,
    ):
        """
        Create call that checks that wait for idle is supported

        string name -- key of the call
        string before -- key of call before which this new call is to be placed
        dict env -- CommandRunner environment variables
        """
        cmd = ["crm_verify"]
        if verbose:
            cmd.extend(["-V", "-V"])
        if cib_tempfile:
            cmd.extend(["--xml-file", cib_tempfile])
        else:
            cmd.append("--live-check")
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stderr=("" if stderr is None else stderr),
                returncode=(0 if stderr is None else 55),
                env=env,
            ),
        )

    def remove_node(
        self,
        node_name,
        stderr="",
        returncode=0,
        name="runner.pcmk.remove_node",
        env=None,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_node", "--force", "--remove", node_name],
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def simulate_cib(
        self,
        new_cib_filepath,
        transitions_filepath,
        cib_xml=None,
        cib_modifiers=None,
        cib_load_name="runner.cib.load",
        stdout="",
        stderr="",
        returncode=0,
        name="runner.pcmk.simulate_cib",
        env=None,
        **modifier_shortcuts,
    ):
        """
        Create a call for simulating effects of cib changes

        string new_cib_filepath -- a temp file for storing a new cib
        string transitions_filepath -- a temp file for storing transitions
        string cib_xml -- cib which will be used for a simulation
        list of callable modifiers -- every callable takes etree.Element and
            returns new etree.Element with desired modification
        string cib_load_name -- key of a call from whose stdout the cib is taken
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        string name -- key of the call
        dict env -- runner environment variables
        dict modifier_shortcuts -- a new modifier is generated from each
            modifier shortcut.
            As key there can be keys of MODIFIER_GENERATORS.
            Value is passed into appropriate generator from MODIFIER_GENERATORS.
            For details see pcs_test.tools.fixture_cib (mainly the variable
            MODIFIER_GENERATORS - please refer it when you are adding params
            here)
        """
        if cib_xml is None:
            cib_xml = modify_cib(
                self.__calls.get(cib_load_name).stdout,
                cib_modifiers,
                **modifier_shortcuts,
            )
        cmd = [
            "crm_simulate",
            "--simulate",
            "--save-output",
            new_cib_filepath,
            "--save-graph",
            transitions_filepath,
            "--xml-pipe",
        ]
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                check_stdin=CheckStdinEqualXml(cib_xml),
                env=env,
            ),
        )

    def get_rule_in_effect_status(
        self,
        rule_id,
        returncode=0,
        name="runner.pcmk.get_rule_in_effect_status",
        cib_load_name="runner.cib.load",
    ):
        """
        Create a call for running a tool to get rule expired status

        string rule_id -- id of the rule to be checked
        int returncode -- result of the check
        sting name -- key of the call
        string cib_load_name -- key of a call from whose stdout the cib is taken
        """
        cib_xml = self.__calls.get(cib_load_name).stdout
        self.__calls.place(
            name,
            RunnerCall(
                ["crm_rule", "--check", "--rule", rule_id, "--xml-text", "-"],
                check_stdin=CheckStdinEqualXml(cib_xml),
                stdout="",
                stderr="",
                returncode=returncode,
            ),
        )

    def resource_agent_self_validation(
        self,
        attributes,
        standard="ocf",
        provider="heartbeat",
        agent_type="Dummy",
        returncode=0,
        output=None,
        stdout="",
        instead=None,
        env=None,
        name="runner.pcmk.resource_agent_self_validation",
    ):
        # pylint: disable=too-many-locals
        if output and stdout:
            raise AssertionError("Cannot specify both output and stdout")
        cmd = [
            "crm_resource",
            "--validate",
            "--output-as",
            "xml",
            "--class",
            standard,
            "--agent",
            agent_type,
        ]
        if provider:
            cmd.extend(["--provider", provider])
        for key, value in sorted(attributes.items()):
            cmd.extend(["--option", f"{key}={value}"])
        if output is not None:
            cmd_str = " ".join(cmd)
            provider_str = f' provider="{provider}"' if provider else ""
            stdout = f"""
            <pacemaker-result api-version="2.15" request="{cmd_str}">
              <resource-agent-action action="validate" class="{standard}" type="{agent_type}"{provider_str}>
                <overrides/>
                <agent-status code="5" message="not installed" execution_code="0" execution_message="complete" reason="environment is invalid, resource considered stopped"/>
                <command code="5">
                  {output}
                </command>
              </resource-agent-action>
              <status code="5" message="Not installed">
                <errors>
                  <error>crm_resource: Error performing operation: Not installed</error>
                </errors>
              </status>
            </pacemaker-result>
            """
        self.__calls.place(
            name,
            RunnerCall(cmd, stdout=stdout, returncode=returncode, env=env),
            instead=instead,
        )

    def stonith_agent_self_validation(
        self,
        attributes,
        agent,
        returncode=0,
        output=None,
        stdout="",
        instead=None,
        env=None,
        name="runner.pcmk.stonith_agent_self_validation",
    ):
        # pylint: disable=too-many-locals
        if output and stdout:
            raise AssertionError("Cannot specify both output and stdout")
        cmd = [
            "stonith_admin",
            "--validate",
            "--output-as",
            "xml",
            "--agent",
            agent,
        ]
        for key, value in sorted(attributes.items()):
            cmd.extend(["--option", f"{key}={value}"])
        if output is not None:
            cmd_str = " ".join(cmd)
            stdout = f"""
            <pacemaker-result api-version="2.22" request="{cmd_str}">
              <validate agent="{agent}" valid="false">
                <command code="-201">
                  {output}
                </command>
              </validate>
              <status code="1" message="Error occurred"/>
            </pacemaker-result>
            """
        self.__calls.place(
            name,
            RunnerCall(cmd, stdout=stdout, returncode=returncode, env=env),
            instead=instead,
        )



from collections import namedtuple

from pcs import settings

Destination = namedtuple("Destination", ["addr", "port"])


class PcsKnownHost(namedtuple("KnownHost", ["name", "token", "dest_list"])):
    @classmethod
    def from_known_host_file_dict(cls, name, known_host_dict):
        dest_list = [
            Destination(conn["addr"], conn["port"])
            for conn in known_host_dict["dest_list"]
        ]
        if not dest_list:
            raise KeyError("no destination defined")
        return cls(name, token=known_host_dict["token"], dest_list=dest_list)

    def to_known_host_dict(self):
        return (
            self.name,
            dict(
                token=self.token,
                dest_list=[
                    dict(
                        addr=dest.addr,
                        port=dest.port,
                    )
                    for dest in self.dest_list
                ],
            ),
        )

    @property
    def dest(self):
        if self.dest_list:
            return self.dest_list[0]
        return Destination(self.name, settings.pcsd_default_port)



import os.path
import re
from typing import (
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    cast,
)

from lxml import etree
from lxml.etree import _Element

from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import join_multilines
from pcs.common.tools import (
    Version,
    format_os_error,
    xml_fromstring,
)
from pcs.common.types import (
    CibRuleInEffectStatus,
    StringCollection,
    StringSequence,
)
from pcs.lib import tools
from pcs.lib.cib.tools import get_pacemaker_version_by_which_cib_was_validated
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker import api_result
from pcs.lib.pacemaker.state import ClusterState
from pcs.lib.resource_agent import ResourceAgentName
from pcs.lib.xml_tools import etree_to_str

__EXITCODE_NOT_CONNECTED = 102
__EXITCODE_CIB_SCOPE_VALID_BUT_NOT_PRESENT = 105
__EXITCODE_WAIT_TIMEOUT = 124
__RESOURCE_REFRESH_OPERATION_COUNT_THRESHOLD = 100


class PacemakerNotConnectedException(LibraryError):
    pass


class FenceHistoryCommandErrorException(Exception):
    pass


### status


def get_cluster_status_xml_raw(runner: CommandRunner) -> Tuple[str, str, int]:
    """
    Run pacemaker tool to get XML status. This function doesn't do any
    processing. Usually, using get_cluster_status_dom is preferred instead.

    runner -- a class for running external processes
    """
    return runner.run(
        [__exec("crm_mon"), "--one-shot", "--inactive", "--output-as", "xml"]
    )


def _get_cluster_status_xml(runner: CommandRunner) -> str:
    """
    Get pacemaker XML status. Using get_cluster_status_dom is preferred instead.

    runner -- a class for running external processes
    """
    stdout, stderr, retval = get_cluster_status_xml_raw(runner)
    if retval == 0:
        return stdout

    # We parse error messages from XML. If we didn't get an XML, we pass it to
    # the exception as a plaintext. If we got an XML but it doesn't conform to
    # the schema, we raise an error.
    try:
        status = _get_status_from_api_result(_get_api_result_dom(stdout))
        message = join_multilines([status.message] + list(status.errors))
    except etree.XMLSyntaxError:
        message = join_multilines([stderr, stdout])
    except etree.DocumentInvalid as e:
        raise LibraryError(
            ReportItem.error(reports.messages.BadClusterStateFormat())
        ) from e
    klass = (
        PacemakerNotConnectedException
        if retval == __EXITCODE_NOT_CONNECTED
        else LibraryError
    )
    raise klass(ReportItem.error(reports.messages.CrmMonError(message)))


def get_cluster_status_dom(runner: CommandRunner) -> _Element:
    try:
        return _get_api_result_dom(_get_cluster_status_xml(runner))
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise LibraryError(
            ReportItem.error(reports.messages.BadClusterStateFormat())
        ) from e


def get_cluster_status_text(
    runner: CommandRunner,
    hide_inactive_resources: bool,
    verbose: bool,
) -> Tuple[str, List[str]]:
    cmd = [__exec("crm_mon"), "--one-shot"]
    if not hide_inactive_resources:
        cmd.append("--inactive")
    if verbose:
        cmd.extend(["--show-detail", "--show-node-attributes", "--failcounts"])
        # by default, pending and failed actions are displayed
        # with verbose==True, we display the whole history
        if is_fence_history_supported_status(runner):
            cmd.append("--fence-history=3")
    stdout, stderr, retval = runner.run(cmd)

    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CrmMonError(join_multilines([stderr, stdout]))
            )
        )
    warnings: List[str] = []
    if stderr.strip():
        warnings = [
            line
            for line in stderr.strip().splitlines()
            if verbose or not line.startswith("DEBUG: ")
        ]

    return stdout.strip(), warnings


def get_ticket_status_text(runner: CommandRunner) -> Tuple[str, str, int]:
    stdout, stderr, retval = runner.run([__exec("crm_ticket"), "--details"])
    return stdout.strip(), stderr.strip(), retval


### cib


def get_cib_xml_cmd_results(
    runner: CommandRunner, scope: Optional[str] = None
) -> tuple[str, str, int]:
    command = [__exec("cibadmin"), "--local", "--query"]
    if scope:
        command.append("--scope={0}".format(scope))
    stdout, stderr, returncode = runner.run(command)
    return stdout, stderr, returncode


def get_cib_xml(runner: CommandRunner, scope: Optional[str] = None) -> str:
    stdout, stderr, retval = get_cib_xml_cmd_results(runner, scope)
    if retval != 0:
        if retval == __EXITCODE_CIB_SCOPE_VALID_BUT_NOT_PRESENT and scope:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.CibLoadErrorScopeMissing(
                        scope, join_multilines([stderr, stdout])
                    )
                )
            )
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibLoadError(join_multilines([stderr, stdout]))
            )
        )
    return stdout


def parse_cib_xml(xml: str) -> _Element:
    return xml_fromstring(xml)


def get_cib(xml: str) -> _Element:
    try:
        return parse_cib_xml(xml)
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise LibraryError(
            ReportItem.error(reports.messages.CibLoadErrorBadFormat(str(e)))
        ) from e


def verify(runner, verbose=False):
    crm_verify_cmd = [__exec("crm_verify")]
    # Currently, crm_verify can suggest up to two -V options but it accepts
    # more than two. We stick with two -V options if verbose mode was enabled.
    if verbose:
        crm_verify_cmd.extend(["-V", "-V"])
    # With the `crm_verify` command it is not possible simply use the
    # environment variable CIB_file because `crm_verify` simply tries to
    # connect to cib file via tool that can fail because: Update does not
    # conform to the configured schema
    # So we use the explicit flag `--xml-file`.
    cib_tmp_file = runner.env_vars.get("CIB_file", None)
    if cib_tmp_file is None:
        crm_verify_cmd.append("--live-check")
    else:
        crm_verify_cmd.extend(["--xml-file", cib_tmp_file])
    stdout, stderr, returncode = runner.run(crm_verify_cmd)
    can_be_more_verbose = False
    if returncode != 0:
        # remove lines with -V options
        rx_v_option = re.compile(r".*-V( -V)* .*more detail.*")
        new_lines = []
        for line in stderr.splitlines(keepends=True):
            if rx_v_option.match(line):
                can_be_more_verbose = True
                continue
            new_lines.append(line)
        # pcs has only one verbose option and cannot be more verbose like
        # `crm_verify` with more -V options. Decision has been made that pcs is
        # limited to only two -V options.
        if verbose:
            can_be_more_verbose = False
        stderr = "".join(new_lines)
    return stdout, stderr, returncode, can_be_more_verbose


def replace_cib_configuration_xml(runner, xml):
    cmd = [
        __exec("cibadmin"),
        "--replace",
        "--verbose",
        "--xml-pipe",
        "--scope",
        "configuration",
    ]
    stdout, stderr, retval = runner.run(cmd, stdin_string=xml)
    if retval != 0:
        raise LibraryError(
            ReportItem.error(reports.messages.CibPushError(stderr, stdout))
        )


def replace_cib_configuration(runner, tree):
    return replace_cib_configuration_xml(runner, etree_to_str(tree))


def push_cib_diff_xml(runner, cib_diff_xml):
    cmd = [
        __exec("cibadmin"),
        "--patch",
        "--verbose",
        "--xml-pipe",
    ]
    stdout, stderr, retval = runner.run(cmd, stdin_string=cib_diff_xml)
    if retval != 0:
        raise LibraryError(
            ReportItem.error(reports.messages.CibPushError(stderr, stdout))
        )


def diff_cibs_xml(
    runner: CommandRunner,
    reporter: ReportProcessor,
    cib_old_xml: str,
    cib_new_xml: str,
) -> str:
    """
    Return xml diff of two CIBs

    runner
    reporter
    cib_old_xml -- original CIB
    cib_new_xml -- modified CIB
    """
    with tools.get_tmp_cib(
        reporter, cib_old_xml
    ) as cib_old_tmp_file, tools.get_tmp_cib(
        reporter, cib_new_xml
    ) as cib_new_tmp_file:
        stdout, stderr, retval = runner.run(
            [
                __exec("crm_diff"),
                "--original",
                cib_old_tmp_file.name,
                "--new",
                cib_new_tmp_file.name,
                "--no-version",
            ]
        )
    #  0 (CRM_EX_OK) - success with no difference
    #  1 (CRM_EX_ERROR) - success with difference
    # 64 (CRM_EX_USAGE) - usage error
    # 65 (CRM_EX_DATAERR) - XML fragments not parseable
    if retval == 0:
        return ""
    if retval > 1:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibDiffError(
                    stderr.strip(), cib_old_xml, cib_new_xml
                )
            )
        )
    return stdout.strip()


def ensure_cib_version(
    runner: CommandRunner,
    cib: _Element,
    version: Version,
    fail_if_version_not_met: bool = True,
) -> Tuple[_Element, bool]:
    """
    Make sure CIB complies to specified schema version (or newer), upgrade CIB
    if necessary. Raise on error. Raise if CIB cannot be upgraded enough to
    meet the required version unless fail_if_version_not_met is set to False.
    Return tuple(upgraded_cib, was_upgraded)

    This method ensures that specified cib is verified by pacemaker with
    version 'version' or newer. If cib doesn't correspond to this version,
    method will try to upgrade cib.
    Returns cib which was verified by pacemaker version 'version' or later.
    Raises LibraryError on any failure.

    runner -- runner
    cib -- cib tree
    version -- required cib version
    fail_if_version_not_met -- allows a 'nice to have' cib upgrade
    """
    version_pre_upgrade = get_pacemaker_version_by_which_cib_was_validated(cib)
    if version_pre_upgrade >= version:
        return cib, False

    _upgrade_cib(runner)
    new_cib_xml = get_cib_xml(runner)

    try:
        new_cib = parse_cib_xml(new_cib_xml)
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise LibraryError(
            ReportItem.error(reports.messages.CibUpgradeFailed(str(e)))
        ) from e

    version_post_upgrade = get_pacemaker_version_by_which_cib_was_validated(
        new_cib
    )
    if version_post_upgrade >= version or not fail_if_version_not_met:
        return new_cib, version_post_upgrade > version_pre_upgrade

    raise LibraryError(
        ReportItem.error(
            reports.messages.CibUpgradeFailedToMinimalRequiredVersion(
                str(version_post_upgrade), str(version)
            )
        )
    )


def _upgrade_cib(runner):
    """
    Upgrade CIB to the latest schema available locally or clusterwise.
    CommandRunner runner
    """
    stdout, stderr, retval = runner.run(
        [__exec("cibadmin"), "--upgrade", "--force"]
    )
    # If we are already on the latest schema available, cibadmin exits with 0.
    # That is fine. We do not know here what version is required anyway. The
    # caller knows that and is responsible for dealing with it.
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibUpgradeFailed(
                    join_multilines([stderr, stdout])
                )
            )
        )


def simulate_cib_xml(runner, cib_xml):
    """
    Run crm_simulate to get effects the cib would have on the live cluster

    CommandRunner runner -- runner
    string cib_xml -- CIB XML to simulate
    """
    try:
        with tools.get_tmp_file() as new_cib_file, tools.get_tmp_file() as transitions_file:
            cmd = [
                __exec("crm_simulate"),
                "--simulate",
                "--save-output",
                new_cib_file.name,
                "--save-graph",
                transitions_file.name,
                "--xml-pipe",
            ]
            stdout, stderr, retval = runner.run(cmd, stdin_string=cib_xml)
            if retval != 0:
                raise LibraryError(
                    ReportItem.error(
                        reports.messages.CibSimulateError(stderr.strip())
                    )
                )
            new_cib_file.seek(0)
            transitions_file.seek(0)
            transitions = transitions_file.read()
            new_cib = new_cib_file.read()
            return stdout, transitions, new_cib
    except OSError as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibSimulateError(format_os_error(e))
            )
        ) from e


def simulate_cib(runner, cib):
    """
    Run crm_simulate to get effects the cib would have on the live cluster

    CommandRunner runner -- runner
    etree cib -- cib tree to simulate
    """
    cib_xml = etree_to_str(cib)
    try:
        plaintext_result, transitions_xml, new_cib_xml = simulate_cib_xml(
            runner, cib_xml
        )
        return (
            plaintext_result.strip(),
            xml_fromstring(transitions_xml),
            xml_fromstring(new_cib_xml),
        )
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise LibraryError(
            ReportItem.error(reports.messages.CibSimulateError(str(e)))
        ) from e


### wait for idle


def wait_for_idle(runner: CommandRunner, timeout: int) -> None:
    """
    Run waiting command. Raise LibraryError if command failed.

    runner -- preconfigured object for running external programs
    timeout -- waiting timeout in seconds, wait indefinitely if non-positive
        integer
    """
    args = [__exec("crm_resource"), "--wait"]
    if timeout > 0:
        args.append("--timeout={0}".format(timeout))
    stdout, stderr, retval = runner.run(args)
    if retval != 0:
        # Usefull info goes to stderr - not only error messages, a list of
        # pending actions in case of timeout goes there as well.
        # We use stdout just to be sure if that's get changed.
        if retval == __EXITCODE_WAIT_TIMEOUT:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(
                        join_multilines([stderr, stdout])
                    )
                )
            )
        raise LibraryError(
            ReportItem.error(
                reports.messages.WaitForIdleError(
                    join_multilines([stderr, stdout])
                )
            )
        )


### nodes


def get_local_node_name(runner):
    stdout, stderr, retval = runner.run([__exec("crm_node"), "--name"])
    if retval != 0:
        klass = (
            PacemakerNotConnectedException
            if retval == __EXITCODE_NOT_CONNECTED
            else LibraryError
        )
        raise klass(
            ReportItem.error(
                reports.messages.PacemakerLocalNodeNameNotFound(
                    join_multilines([stderr, stdout])
                )
            )
        )
    return stdout.strip()


def get_local_node_status(runner):
    try:
        cluster_status = ClusterState(get_cluster_status_dom(runner))
        node_name = get_local_node_name(runner)
    except PacemakerNotConnectedException:
        return {"offline": True}
    for node_status in cluster_status.node_section.nodes:
        if node_status.attrs.name == node_name:
            result = {
                "offline": False,
            }
            for attr in (
                "id",
                "name",
                "type",
                "online",
                "standby",
                "standby_onfail",
                "maintenance",
                "pending",
                "unclean",
                "shutdown",
                "expected_up",
                "is_dc",
                "resources_running",
            ):
                result[attr] = getattr(node_status.attrs, attr)
            return result
    raise LibraryError(
        ReportItem.error(reports.messages.NodeNotFound(node_name))
    )


def remove_node(runner, node_name):
    stdout, stderr, retval = runner.run(
        [
            __exec("crm_node"),
            "--force",
            "--remove",
            node_name,
        ]
    )
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.NodeRemoveInPacemakerFailed(
                    node_list_to_remove=[node_name],
                    reason=join_multilines([stderr, stdout]),
                )
            )
        )


### resources


def resource_cleanup(
    runner: CommandRunner,
    resource: Optional[str] = None,
    node: Optional[str] = None,
    operation: Optional[str] = None,
    interval: Optional[str] = None,
    strict: bool = False,
):
    cmd = [__exec("crm_resource"), "--cleanup"]
    if resource:
        cmd.extend(["--resource", resource])
    if node:
        cmd.extend(["--node", node])
    if operation:
        cmd.extend(["--operation", operation])
    if interval:
        cmd.extend(["--interval", interval])
    if strict:
        cmd.extend(["--force"])

    stdout, stderr, retval = runner.run(cmd)

    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.ResourceCleanupError(
                    join_multilines([stderr, stdout]), resource, node
                )
            )
        )
    # usefull output (what has been done) goes to stderr
    return join_multilines([stdout, stderr])


def resource_refresh(
    runner: CommandRunner,
    resource: Optional[str] = None,
    node: Optional[str] = None,
    strict: bool = False,
    force: bool = False,
):
    if not force and not node and not resource:
        summary = ClusterState(get_cluster_status_dom(runner)).summary
        operations = summary.nodes.attrs.count * summary.resources.attrs.count
        if operations > __RESOURCE_REFRESH_OPERATION_COUNT_THRESHOLD:
            raise LibraryError(
                ReportItem(
                    reports.item.ReportItemSeverity.error(reports.codes.FORCE),
                    reports.messages.ResourceRefreshTooTimeConsuming(
                        __RESOURCE_REFRESH_OPERATION_COUNT_THRESHOLD
                    ),
                )
            )

    cmd = [__exec("crm_resource"), "--refresh"]
    if resource:
        cmd.extend(["--resource", resource])
    if node:
        cmd.extend(["--node", node])
    if strict:
        cmd.extend(["--force"])

    stdout, stderr, retval = runner.run(cmd)

    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.ResourceRefreshError(
                    join_multilines([stderr, stdout]), resource, node
                )
            )
        )
    # usefull output (what has been done) goes to stderr
    return join_multilines([stdout, stderr])


def resource_move(runner, resource_id, node=None, master=False, lifetime=None):
    return _resource_move_ban_clear(
        runner,
        "--move",
        resource_id,
        node=node,
        master=master,
        lifetime=lifetime,
    )


def resource_ban(runner, resource_id, node=None, master=False, lifetime=None):
    return _resource_move_ban_clear(
        runner,
        "--ban",
        resource_id,
        node=node,
        master=master,
        lifetime=lifetime,
    )


def resource_unmove_unban(
    runner, resource_id, node=None, master=False, expired=False
):
    return _resource_move_ban_clear(
        runner,
        "--clear",
        resource_id,
        node=node,
        master=master,
        expired=expired,
    )


def has_resource_unmove_unban_expired_support(runner):
    return _is_in_pcmk_tool_help(runner, "crm_resource", ["--expired"])


def _resource_move_ban_clear(
    runner,
    action,
    resource_id,
    node=None,
    master=False,
    lifetime=None,
    expired=False,
):
    command = [
        __exec("crm_resource"),
        action,
        "--resource",
        resource_id,
    ]
    if node:
        command.extend(["--node", node])
    if master:
        command.extend(["--master"])
    if lifetime:
        command.extend(["--lifetime", lifetime])
    if expired:
        command.extend(["--expired"])
    stdout, stderr, retval = runner.run(command)
    return stdout, stderr, retval


### fence history


def is_fence_history_supported_status(runner: CommandRunner) -> bool:
    return _is_in_pcmk_tool_help(runner, "crm_mon", ["--fence-history"])


def is_fence_history_supported_management(runner: CommandRunner) -> bool:
    return _is_in_pcmk_tool_help(
        runner, "stonith_admin", ["--history", "--broadcast", "--cleanup"]
    )


def fence_history_cleanup(runner, node=None):
    return _run_fence_history_command(runner, "--cleanup", node)


def fence_history_text(runner, node=None):
    return _run_fence_history_command(runner, "--verbose", node)


def fence_history_update(runner):
    # Pacemaker always prints "gather fencing-history from all nodes" even if a
    # node is specified. However, --history expects a value, so we must provide
    # it. Otherwise "--broadcast" would be considered a value of "--history".
    return _run_fence_history_command(runner, "--broadcast", node=None)


def _run_fence_history_command(runner, command, node=None):
    stdout, stderr, retval = runner.run(
        [__exec("stonith_admin"), "--history", node if node else "*", command]
    )
    if retval != 0:
        raise FenceHistoryCommandErrorException(
            join_multilines([stderr, stdout])
        )
    return stdout.strip()


### tools


def has_rule_in_effect_status_tool() -> bool:
    return os.path.isfile(__exec("crm_rule"))


def get_rule_in_effect_status(
    runner: CommandRunner, cib_xml: str, rule_id: str
) -> CibRuleInEffectStatus:
    """
    Figure out if a rule is in effect, expired or not yet in effect

    runner -- a class for running external processes
    cib_xml -- CIB containing rules
    rule_id -- ID of the rule to be checked
    """
    # TODO Once crm_rule is capable of evaluating more than one rule per go, we
    # should make use of it. Running the tool for each rule may really slow pcs
    # down.
    translation_map = {
        0: CibRuleInEffectStatus.IN_EFFECT,
        110: CibRuleInEffectStatus.EXPIRED,
        111: CibRuleInEffectStatus.NOT_YET_IN_EFFECT,
        # 105:non-existent
        # 112: undetermined (rule is too complicated for current implementation)
    }
    dummy_stdout, dummy_stderr, retval = runner.run(
        [__exec("crm_rule"), "--check", "--rule", rule_id, "--xml-text", "-"],
        stdin_string=cib_xml,
    )
    return translation_map.get(retval, CibRuleInEffectStatus.UNKNOWN)


def _get_api_result_dom(xml: str) -> _Element:
    # raises etree.XMLSyntaxError and etree.DocumentInvalid
    rng = settings.pacemaker_api_result_schema
    dom = xml_fromstring(xml)
    if os.path.isfile(rng):
        etree.RelaxNG(file=rng).assertValid(dom)
    return dom


def _get_status_from_api_result(dom: _Element) -> api_result.Status:
    errors = []
    status_el = cast(_Element, dom.find("./status"))
    errors_el = status_el.find("errors")
    if errors_el is not None:
        errors = [
            str((error_el.text or "")).strip()
            for error_el in errors_el.iterfind("error")
        ]
    return api_result.Status(
        code=int(str(status_el.get("code"))),
        message=str(status_el.get("message")),
        errors=errors,
    )


# shortcut for getting a full path to a pacemaker executable
def __exec(name):
    return os.path.join(settings.pacemaker_binaries, name)


def _is_in_pcmk_tool_help(
    runner: CommandRunner, tool: str, text_list: StringCollection
) -> bool:
    stdout, stderr, dummy_retval = runner.run([__exec(tool), "--help-all"])
    # Help goes to stderr but we check stdout as well if that gets changed. Use
    # generators in all to return early.
    return all(text in stderr for text in text_list) or all(
        text in stdout for text in text_list
    )


def is_getting_resource_digest_supported(runner):
    return _is_in_pcmk_tool_help(runner, "crm_resource", "--digests")


def get_resource_digests(
    runner: CommandRunner,
    resource_id: str,
    node_name: str,
    resource_options: Dict[str, str],
    crm_meta_attributes: Optional[Dict[str, Optional[str]]] = None,
) -> Dict[str, Optional[str]]:
    """
    Get set of digests for a resource using crm_resource utility. There are 3
    types of digests: all, nonreloadable and nonprivate. Resource can have one
    or more digests types depending on the resource parameters.

    runner -- command runner instance
    resource_id -- resource id
    node_name -- name of the node where resource is running
    resource_options -- resource options with updated values
    crm_meta_attributes -- parameters of a monitor operation
    """
    # pylint: disable=too-many-locals
    if crm_meta_attributes is None:
        crm_meta_attributes = {}
    command = [
        __exec("crm_resource"),
        "--digests",
        "--resource",
        resource_id,
        "--node",
        node_name,
        "--output-as",
        "xml",
        *[f"{key}={value}" for key, value in resource_options.items()],
        *[
            f"CRM_meta_{key}={value}"
            for key, value in crm_meta_attributes.items()
            if value is not None
        ],
    ]
    stdout, stderr, retval = runner.run(command)

    def error_exception(message):
        return LibraryError(
            ReportItem.error(
                reports.messages.UnableToGetResourceOperationDigests(message)
            )
        )

    try:
        dom = _get_api_result_dom(stdout)
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise error_exception(join_multilines([stderr, stdout])) from e

    if retval != 0:
        status = _get_status_from_api_result(dom)
        raise error_exception(
            join_multilines([status.message] + list(status.errors))
        )

    digests = {}
    for digest_type in ["all", "nonprivate", "nonreloadable"]:
        xpath_result = cast(
            List[str],
            dom.xpath(
                "./digests/digest[@type=$digest_type]/@hash",
                digest_type=digest_type,
            ),
        )
        digests[digest_type] = xpath_result[0] if xpath_result else None
    if not any(digests.values()):
        raise error_exception(join_multilines([stderr, stdout]))
    return digests


def _validate_stonith_instance_attributes_via_pcmk(
    cmd_runner: CommandRunner,
    agent_name: ResourceAgentName,
    instance_attributes: Mapping[str, str],
) -> tuple[Optional[bool], str]:
    cmd = [
        settings.stonith_admin,
        "--validate",
        "--output-as",
        "xml",
        "--agent",
        agent_name.type,
    ]
    return _handle_instance_attributes_validation_via_pcmk(
        cmd_runner,
        cmd,
        "./validate/command/output",
        instance_attributes,
    )


def _validate_resource_instance_attributes_via_pcmk(
    cmd_runner: CommandRunner,
    agent_name: ResourceAgentName,
    instance_attributes: Mapping[str, str],
) -> tuple[Optional[bool], str]:
    cmd = [
        settings.crm_resource_binary,
        "--validate",
        "--output-as",
        "xml",
        "--class",
        agent_name.standard,
        "--agent",
        agent_name.type,
    ]
    if agent_name.provider:
        cmd.extend(["--provider", agent_name.provider])
    return _handle_instance_attributes_validation_via_pcmk(
        cmd_runner,
        cmd,
        "./resource-agent-action/command/output",
        instance_attributes,
    )


def _handle_instance_attributes_validation_via_pcmk(
    cmd_runner: CommandRunner,
    cmd: StringSequence,
    data_xpath: str,
    instance_attributes: Mapping[str, str],
) -> tuple[Optional[bool], str]:
    full_cmd = list(cmd)
    for key, value in sorted(instance_attributes.items()):
        full_cmd.extend(["--option", f"{key}={value}"])
    stdout, dummy_stderr, return_value = cmd_runner.run(full_cmd)
    try:
        # dom = _get_api_result_dom(stdout)
        dom = xml_fromstring(stdout)
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        return None, str(e)
    result = "\n".join(
        "\n".join(
            line.strip() for line in item.text.split("\n") if line.strip()
        )
        for item in dom.iterfind(data_xpath)
        if item.get("source") == "stderr" and item.text
    ).strip()
    return return_value == 0, result


def validate_resource_instance_attributes_via_pcmk(
    cmd_runner: CommandRunner,
    resource_agent_name: ResourceAgentName,
    instance_attributes: Mapping[str, str],
) -> tuple[Optional[bool], str]:
    if resource_agent_name.is_stonith:
        return _validate_stonith_instance_attributes_via_pcmk(
            cmd_runner,
            resource_agent_name,
            instance_attributes,
        )
    return _validate_resource_instance_attributes_via_pcmk(
        cmd_runner,
        resource_agent_name,
        instance_attributes,
    )



import os.path
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.file import RawFileError
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib.corosync import live as lib
from pcs.lib.external import CommandRunner

from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.misc import get_test_resource as rc

# pylint: disable=no-self-use


class GetLocalCorosyncConfTest(TestCase):
    def test_success(self):
        path = rc("corosync.conf")
        settings.corosync_conf_file = path
        with open(path) as a_file:
            self.assertEqual(lib.get_local_corosync_conf(), a_file.read())

    def test_error(self):
        path = rc("corosync.conf.nonexistent")
        settings.corosync_conf_file = path
        assert_raise_library_error(
            lib.get_local_corosync_conf,
            (
                severity.ERROR,
                report_codes.FILE_IO_ERROR,
                {
                    "file_type_code": file_type_codes.COROSYNC_CONF,
                    "operation": RawFileError.ACTION_READ,
                    "reason": f"No such file or directory: '{path}'",
                    "file_path": path,
                },
            ),
        )


class GetQuorumStatusTextTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)
        self.quorum_tool = os.path.join(
            settings.corosync_binaries, "corosync-quorumtool"
        )

    def test_success(self):
        self.mock_runner.run.return_value = ("status info", "", 0)
        self.assertEqual(
            "status info", lib.get_quorum_status_text(self.mock_runner)
        )
        self.mock_runner.run.assert_called_once_with([self.quorum_tool, "-p"])

    def test_success_with_retval_1(self):
        self.mock_runner.run.return_value = ("status info", "", 1)
        self.assertEqual(
            "status info", lib.get_quorum_status_text(self.mock_runner)
        )
        self.mock_runner.run.assert_called_once_with([self.quorum_tool, "-p"])

    def test_error(self):
        self.mock_runner.run.return_value = ("some info", "status error", 2)
        with self.assertRaises(lib.QuorumStatusReadException) as cm:
            lib.get_quorum_status_text(self.mock_runner)
        self.mock_runner.run.assert_called_once_with([self.quorum_tool, "-p"])
        self.assertEqual(cm.exception.reason, "status error")


class SetExpectedVotesTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def path(self, name):
        return os.path.join(settings.corosync_binaries, name)

    def test_success(self):
        cmd_retval = 0
        cmd_stdout = "cmd output"
        cmd_stderr = ""
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        mock_runner.run.return_value = (cmd_stdout, cmd_stderr, cmd_retval)

        lib.set_expected_votes(mock_runner, 3)

        mock_runner.run.assert_called_once_with(
            [self.path("corosync-quorumtool"), "-e", "3"]
        )

    def test_error(self):
        cmd_retval = 1
        cmd_stdout = "cmd output"
        cmd_stderr = "cmd stderr"
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        mock_runner.run.return_value = (cmd_stdout, cmd_stderr, cmd_retval)

        assert_raise_library_error(
            lambda: lib.set_expected_votes(mock_runner, 3),
            (
                severity.ERROR,
                report_codes.COROSYNC_QUORUM_SET_EXPECTED_VOTES_ERROR,
                {
                    "reason": cmd_stderr,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [self.path("corosync-quorumtool"), "-e", "3"]
        )


class QuorumStatusParse(TestCase):
    def test_quorate_no_qdevice(self):
        facade = lib.QuorumStatusFacade.from_string(
            dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:28 2015
                Quorum provider:  corosync_votequorum
                Nodes:            3
                Node ID:          1
                Ring ID:          19860
                Quorate:          Yes

                Votequorum information
                ----------------------
                Expected votes:   3
                Highest expected: 3
                Total votes:      3
                Quorum:           2
                Flags:            Quorate

                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                         1          3         NR rh70-node1
                         2          2         NR rh70-node2 (local)
                         3          1         NR rh70-node3
            """
            )
        )
        self.assertEqual(facade.is_quorate, True)
        self.assertEqual(facade.votes_needed_for_quorum, 2)
        self.assertEqual(facade.qdevice_votes, 0)
        self.assertEqual(
            facade.node_list,
            [
                lib.QuorumStatusNode(name="rh70-node1", votes=3, local=False),
                lib.QuorumStatusNode(name="rh70-node2", votes=2, local=True),
                lib.QuorumStatusNode(name="rh70-node3", votes=1, local=False),
            ],
        )
        self.assertEqual(facade.qdevice_list, [])

    def test_quorate_with_qdevice(self):
        facade = lib.QuorumStatusFacade.from_string(
            dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:28 2015
                Quorum provider:  corosync_votequorum
                Nodes:            3
                Node ID:          1
                Ring ID:          19860
                Quorate:          Yes

                Votequorum information
                ----------------------
                Expected votes:   10
                Highest expected: 10
                Total votes:      10
                Quorum:           6
                Flags:            Quorate Qdevice

                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                         1          3    A,V,MNW rh70-node1
                         2          2    A,V,MNW rh70-node2 (local)
                         3          1    A,V,MNW rh70-node3
                         0          4            Qdevice
            """
            )
        )
        self.assertEqual(facade.is_quorate, True)
        self.assertEqual(facade.votes_needed_for_quorum, 6)
        self.assertEqual(facade.qdevice_votes, 4)
        self.assertEqual(
            facade.node_list,
            [
                lib.QuorumStatusNode(name="rh70-node1", votes=3, local=False),
                lib.QuorumStatusNode(name="rh70-node2", votes=2, local=True),
                lib.QuorumStatusNode(name="rh70-node3", votes=1, local=False),
            ],
        )
        self.assertEqual(
            facade.qdevice_list,
            [
                lib.QuorumStatusNode(name="Qdevice", votes=4, local=False),
            ],
        )

    def test_quorate_with_qdevice_lost(self):
        facade = lib.QuorumStatusFacade.from_string(
            dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:28 2015
                Quorum provider:  corosync_votequorum
                Nodes:            3
                Node ID:          1
                Ring ID:          19860
                Quorate:          Yes

                Votequorum information
                ----------------------
                Expected votes:   10
                Highest expected: 10
                Total votes:      6
                Quorum:           6
                Flags:            Quorate Qdevice

                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                         1          3   NA,V,MNW rh70-node1
                         2          2   NA,V,MNW rh70-node2 (local)
                         3          1   NA,V,MNW rh70-node3
                         0          0            Qdevice (votes 4)
            """
            )
        )
        self.assertEqual(facade.is_quorate, True)
        self.assertEqual(facade.votes_needed_for_quorum, 6)
        self.assertEqual(facade.qdevice_votes, 0)
        self.assertEqual(
            facade.node_list,
            [
                lib.QuorumStatusNode(name="rh70-node1", votes=3, local=False),
                lib.QuorumStatusNode(name="rh70-node2", votes=2, local=True),
                lib.QuorumStatusNode(name="rh70-node3", votes=1, local=False),
            ],
        )
        self.assertEqual(
            facade.qdevice_list,
            [
                lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
            ],
        )

    def test_no_quorate_no_qdevice(self):
        facade = lib.QuorumStatusFacade.from_string(
            dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:35 2015
                Quorum provider:  corosync_votequorum
                Nodes:            1
                Node ID:          1
                Ring ID:          19868
                Quorate:          No

                Votequorum information
                ----------------------
                Expected votes:   3
                Highest expected: 3
                Total votes:      1
                Quorum:           2 Activity blocked
                Flags:            

                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                             1          1         NR rh70-node1 (local)
            """
            )
        )
        self.assertEqual(facade.is_quorate, False)
        self.assertEqual(facade.votes_needed_for_quorum, 2)
        self.assertEqual(facade.qdevice_votes, 0)
        self.assertEqual(
            facade.node_list,
            [
                lib.QuorumStatusNode(name="rh70-node1", votes=1, local=True),
            ],
        )
        self.assertEqual(facade.qdevice_list, [])

    def test_no_quorate_with_qdevice(self):
        facade = lib.QuorumStatusFacade.from_string(
            dedent(
                """\
                Quorum information
                ------------------
                Date:             Fri Jan 16 13:03:35 2015
                Quorum provider:  corosync_votequorum
                Nodes:            1
                Node ID:          1
                Ring ID:          19868
                Quorate:          No

                Votequorum information
                ----------------------
                Expected votes:   3
                Highest expected: 3
                Total votes:      1
                Quorum:           2 Activity blocked
                Flags:            Qdevice

                Membership information
                ----------------------
                    Nodeid      Votes    Qdevice Name
                         1          1         NR rh70-node1 (local)
                         0          0            Qdevice (votes 1)
            """
            )
        )
        self.assertEqual(facade.is_quorate, False)
        self.assertEqual(facade.votes_needed_for_quorum, 2)
        self.assertEqual(facade.qdevice_votes, 0)
        self.assertEqual(
            facade.node_list,
            [
                lib.QuorumStatusNode(name="rh70-node1", votes=1, local=True),
            ],
        )
        self.assertEqual(
            facade.qdevice_list,
            [
                lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
            ],
        )

    def test_error_empty_string(self):
        with self.assertRaises(lib.QuorumStatusParsingException) as cm:
            lib.QuorumStatusFacade.from_string("")
        self.assertEqual(
            cm.exception.reason,
            "Missing required section(s): 'node_list', 'quorate', 'quorum'",
        )

    def test_error_missing_quorum(self):
        with self.assertRaises(lib.QuorumStatusParsingException) as cm:
            lib.QuorumStatusFacade.from_string(
                dedent(
                    """\
                    Quorum information
                    ------------------
                    Date:             Fri Jan 16 13:03:28 2015
                    Quorum provider:  corosync_votequorum
                    Nodes:            3
                    Node ID:          1
                    Ring ID:          19860
                    Quorate:          Yes

                    Votequorum information
                    ----------------------
                    Expected votes:   3
                    Highest expected: 3
                    Total votes:      3
                    Quorum:           
                    Flags:            Quorate

                    Membership information
                    ----------------------
                        Nodeid      Votes    Qdevice Name
                             1          1         NR rh70-node1 (local)
                             2          1         NR rh70-node2
                             3          1         NR rh70-node3
                """
                )
            )
        self.assertEqual(
            cm.exception.reason,
            "Unable to read number of votes needed for quorum",
        )

    def test_error_quorum_garbage(self):
        with self.assertRaises(lib.QuorumStatusParsingException) as cm:
            lib.QuorumStatusFacade.from_string(
                dedent(
                    """\
                    Quorum information
                    ------------------
                    Date:             Fri Jan 16 13:03:28 2015
                    Quorum provider:  corosync_votequorum
                    Nodes:            3
                    Node ID:          1
                    Ring ID:          19860
                    Quorate:          Yes

                    Votequorum information
                    ----------------------
                    Expected votes:   3
                    Highest expected: 3
                    Total votes:      3
                    Quorum:           Foo
                    Flags:            Quorate

                    Membership information
                    ----------------------
                        Nodeid      Votes    Qdevice Name
                             1          1         NR rh70-node1 (local)
                             2          1         NR rh70-node2
                             3          1         NR rh70-node3
                """
                )
            )
        self.assertEqual(
            cm.exception.reason,
            "Unable to read number of votes needed for quorum",
        )

    def test_error_node_votes_garbage(self):
        with self.assertRaises(lib.QuorumStatusParsingException) as cm:
            lib.QuorumStatusFacade.from_string(
                dedent(
                    """\
                    Quorum information
                    ------------------
                    Date:             Fri Jan 16 13:03:28 2015
                    Quorum provider:  corosync_votequorum
                    Nodes:            3
                    Node ID:          1
                    Ring ID:          19860
                    Quorate:          Yes

                    Votequorum information
                    ----------------------
                    Expected votes:   3
                    Highest expected: 3
                    Total votes:      3
                    Quorum:           2
                    Flags:            Quorate

                    Membership information
                    ----------------------
                        Nodeid      Votes    Qdevice Name
                             1          1         NR rh70-node1 (local)
                             2        foo         NR rh70-node2
                             3          1         NR rh70-node3
                """
                )
            )
        self.assertEqual(cm.exception.reason, "")


class QuorumStatusQuorumLossNodes(TestCase):
    def test_not_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=False,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node3"]), False
        )

    def test_one_node_still_quorate_1(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node3"]), False
        )

    def test_one_node_still_quorate_2(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2"]), False
        )

    def test_one_node_quorum_loss(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node1"]), True
        )

    def test_more_nodes_still_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=4, local=True),
                    lib.QuorumStatusNode(name="node2", votes=1, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2", "node3"]),
            False,
        )

    def test_more_nodes_quorum_loss(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2", "node3"]),
            True,
        )

    def test_qdevice_still_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=3,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=1, local=True),
                    lib.QuorumStatusNode(name="node2", votes=1, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=1, local=False),
                ],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2"]), False
        )

    def test_qdevice_quorum_lost(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=3,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=1, local=True),
                    lib.QuorumStatusNode(name="node2", votes=1, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=1, local=False),
                ],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2", "node3"]),
            True,
        )

    def test_qdevice_lost_still_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,  # expect qdevice votes == 1
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=2, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=2, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
                ],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2"]), False
        )

    def test_qdevice_lost_quorum_lost(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,  # expect qdevice votes == 1
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=2, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=2, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
                ],
            )
        )
        self.assertEqual(
            facade.stopping_nodes_cause_quorum_loss(["node2", "node3"]),
            True,
        )


class QuorumStatusQuorumLossLocal(TestCase):
    def test_not_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=False,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_local_node_not_in_list(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=1,
                node_list=[
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_local_node_alone_in_list(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=1,
                node_list=[
                    lib.QuorumStatusNode(name="node3", votes=1, local=True),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), True)

    def test_local_node_still_quorate_1(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=False),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=True),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_local_node_still_quorate_2(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=False),
                    lib.QuorumStatusNode(name="node2", votes=2, local=True),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_local_node_quorum_loss(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=3, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), True)

    def test_qdevice_still_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=3,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=1, local=True),
                    lib.QuorumStatusNode(name="node2", votes=1, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=1, local=False),
                ],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_qdevice_quorum_lost(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=4, local=True),
                    lib.QuorumStatusNode(name="node2", votes=1, local=False),
                    lib.QuorumStatusNode(name="node3", votes=1, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=1, local=False),
                ],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), True)

    def test_qdevice_lost_still_quorate(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=4,  # expect qdevice votes == 1
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=2, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=2, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
                ],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), False)

    def test_qdevice_lost_quorum_lost(self):
        facade = lib.QuorumStatusFacade(
            lib.QuorumStatus(
                is_quorate=True,
                votes_needed_for_quorum=5,  # expect qdevice votes == 1
                node_list=[
                    lib.QuorumStatusNode(name="node1", votes=4, local=True),
                    lib.QuorumStatusNode(name="node2", votes=2, local=False),
                    lib.QuorumStatusNode(name="node3", votes=2, local=False),
                ],
                qdevice_list=[
                    lib.QuorumStatusNode(name="Qdevice", votes=0, local=False),
                ],
            )
        )
        self.assertEqual(facade.stopping_local_node_cause_quorum_loss(), True)



import os.path
from unittest import TestCase

from lxml import etree

from pcs_test import PROJECT_ROOT


class TestCapabilities(TestCase):
    def test_parsable(self):
        # pylint: disable=no-self-use
        capabilities_dir = os.path.join(PROJECT_ROOT, "pcsd")
        dom = etree.parse(os.path.join(capabilities_dir, "capabilities.xml"))
        etree.RelaxNG(
            file=os.path.join(capabilities_dir, "capabilities.rng")
        ).assertValid(dom)



import json

from pcs import utils
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.reports.output import deprecation_warning
from pcs.lib.commands import cluster_property


def list_property_deprecated(lib, argv, modifiers):
    deprecation_warning(
        "This command is deprecated and will be removed. "
        "Please use 'pcs property config' instead."
    )
    return list_property(lib, argv, modifiers)


def list_property(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --all - list all properties
      * --defaults - list only default values of properties
    """
    # pylint: disable=too-many-branches
    del lib
    modifiers.ensure_only_supported(
        "--defaults",
        "--all",
        "-f",
    )
    if len(argv) > 1:
        raise CmdLineInputError()
    print_all = len(argv) == 0

    if modifiers.is_specified("--all") and modifiers.is_specified("--defaults"):
        utils.err("you cannot specify both --all and --defaults")

    if modifiers.get("--all") or modifiers.get("--defaults"):
        if not print_all:
            utils.err(
                "you cannot specify a property when using --all or --defaults"
            )
        properties = get_default_properties()
    else:
        properties = {}

    if not modifiers.get("--defaults"):
        configured_properties = utils.get_set_properties(
            None if print_all else argv[0], properties
        )
        if modifiers.get("--all"):
            properties.update(configured_properties)
        else:
            properties = configured_properties

    if properties:
        print("Cluster Properties:")
        for prop, val in sorted(properties.items()):
            print(" {0}: {1}".format(prop, val))
    else:
        properties = get_default_properties()
        print("Cluster Properties:")
        if not print_all:
            if argv[0] in properties:
                print(" {0}: {1}".format(argv[0], properties[argv[0]]))


def get_default_properties():
    """
    Commandline options: no options
    """
    parameters = {}
    lib_env = utils.get_lib_env()
    prop_def_dict = cluster_property.get_cluster_properties_definition_legacy(
        lib_env
    )
    for name, prop in prop_def_dict.items():
        parameters[name] = prop["default"]
    return parameters


def print_cluster_properties_definition(lib, argv, modifiers):
    """
    Options: no options
    """
    del lib
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    lib_env = utils.get_lib_env()
    prop_def_dict = cluster_property.get_cluster_properties_definition_legacy(
        lib_env
    )
    print(json.dumps(prop_def_dict))



from enum import Enum


class PermissionTargetType(str, Enum):
    USER = "user"
    GROUP = "group"


class PermissionAccessType(str, Enum):
    READ = "read"
    WRITE = "write"
    GRANT = "grant"
    FULL = "full"
    SUPERUSER = "superuser"



import grp
import pwd

from pcs.common.tools import StringCollection

from .types import (
    AuthUser,
    DesiredUser,
)


class UserGroupsError(Exception):
    pass


def get_user_groups(username: str) -> list[str]:
    try:
        return [
            group.gr_name
            for group in grp.getgrall()
            if username in group.gr_mem
        ] + [
            # `group.gr_mem` does not contain the `username` when `group` is
            # primary for the `username` (the same is in /etc/group). So it is
            # necessary to add the primary group.
            grp.getgrgid(pwd.getpwnam(username).pw_gid).gr_name
        ]
    except KeyError as e:
        raise UserGroupsError from e


def get_effective_user(
    authenticated_user: AuthUser,
    effective_user_candidate: DesiredUser,
) -> AuthUser:
    username = authenticated_user.username
    groups: StringCollection = authenticated_user.groups
    if effective_user_candidate.username:
        username = effective_user_candidate.username
        if effective_user_candidate.groups:
            groups = effective_user_candidate.groups
    return AuthUser(username=username, groups=tuple(groups))



from pcs_test.tools.command_env.mock_runner import Call as RunnerCall
from pcs_test.tools.command_env.mock_runner import CheckStdinEqualXml
from pcs_test.tools.fixture_cib import modify_cib
from pcs_test.tools.misc import get_test_resource as rc

CIB_FILENAME = "cib-empty.xml"


class CibShortcuts:
    def __init__(self, calls):
        """
        CallCollection calls -- provides access to call list
        """
        self.__calls = calls
        self.cib_filename = CIB_FILENAME

    def load(
        self,
        modifiers=None,
        name="runner.cib.load",
        filename=None,
        before=None,
        returncode=0,
        stderr=None,
        instead=None,
        env=None,
        **modifier_shortcuts,
    ):
        """
        Create call for loading cib.

        string name -- key of the call
        list of callable modifiers -- every callable takes etree.Element and
            returns new etree.Element with desired modification.
        string filename -- points to file with cib in the content
        string before -- key of call before which this new call is to be placed
        int returncode
        string stderr
        string instead -- key of call instead of which this new call is to be
            placed
        dict env -- CommandRunner environment variables
        dict modifier_shortcuts -- a new modifier is generated from each
            modifier shortcut.
            As key there can be keys of MODIFIER_GENERATORS.
            Value is passed into appropriate generator from MODIFIER_GENERATORS.
            For details see pcs_test.tools.fixture_cib (mainly the variable
            MODIFIER_GENERATORS - please refer it when you are adding params
            here)
        """
        # pylint: disable=too-many-arguments
        if (returncode != 0 or stderr is not None) and (
            modifiers is not None or filename is not None or modifier_shortcuts
        ):
            raise AssertionError(
                "Do not combine parameters 'returncode' and 'stderr' with"
                " parameters 'modifiers', 'filename' and 'modifier_shortcuts'"
            )

        command = ["cibadmin", "--local", "--query"]
        if returncode != 0:
            call = RunnerCall(
                command, stderr=stderr, returncode=returncode, env=env
            )
        else:
            with open(
                rc(filename if filename else self.cib_filename),
            ) as cib_file:
                cib = modify_cib(
                    cib_file.read(), modifiers, **modifier_shortcuts
                )
                call = RunnerCall(command, stdout=cib, env=env)

        self.__calls.place(name, call, before=before, instead=instead)

    def load_content(
        self,
        cib,
        returncode=0,
        stderr=None,
        name="runner.cib.load_content",
        instead=None,
        before=None,
        env=None,
    ):
        """
        Create call for loading CIB specified by its full content

        string cib -- CIB data (stdout of the loading process)
        string stderr -- error returned from the loading process
        int returncode -- exit code of the loading process
        string name -- key of the call
        string instead -- key of call instead of which this new call is to be
            placed
        string before -- key of call before which this new call is to be placed
        dict env -- CommandRunner environment variables
        """
        command = ["cibadmin", "--local", "--query"]
        if returncode != 0:
            call = RunnerCall(
                command, stderr=stderr, returncode=returncode, env=env
            )
        else:
            call = RunnerCall(command, stdout=cib, env=env)
        self.__calls.place(name, call, before=before, instead=instead)

    def push(
        self,
        modifiers=None,
        name="runner.cib.push",
        load_key="runner.cib.load",
        instead=None,
        stderr="",
        returncode=0,
        **modifier_shortcuts,
    ):
        """
        Create call for pushing cib.
        Cib is taken from the load call by default.

        string name -- key of the call
        list of callable modifiers -- every callable takes etree.Element and
            returns new etree.Element with desired modification.
        string load_key -- key of a call from which stdout can be cib taken
        string instead -- key of call instead of which this new call is to be
            placed
        dict modifier_shortcuts -- a new modifier is generated from each
            modifier shortcut.
            As key there can be keys of MODIFIER_GENERATORS.
            Value is passed into appropriate generator from MODIFIER_GENERATORS.
            For details see pcs_test.tools.fixture_cib (mainly the variable
            MODIFIER_GENERATORS - please refer it when you are adding params
            here)
        """
        cib = modify_cib(
            self.__calls.get(load_key).stdout, modifiers, **modifier_shortcuts
        )
        self.__calls.place(
            name,
            RunnerCall(
                [
                    "cibadmin",
                    "--replace",
                    "--verbose",
                    "--xml-pipe",
                    "--scope",
                    "configuration",
                ],
                stderr=stderr,
                returncode=returncode,
                check_stdin=CheckStdinEqualXml(cib),
            ),
            instead=instead,
        )

    def push_independent(
        self,
        cib,
        name="runner.cib.push_independent",
        instead=None,
    ):
        """
        Create call for pushing cib.
        Cib is specified as an argument.

        string name -- key of the call
        string cib -- whole cib to push
        string instead -- key of call instead of which this new call is to be
            placed
        """
        self.__calls.place(
            name,
            RunnerCall(
                [
                    "cibadmin",
                    "--replace",
                    "--verbose",
                    "--xml-pipe",
                    "--scope",
                    "configuration",
                ],
                check_stdin=CheckStdinEqualXml(cib),
            ),
            instead=instead,
        )

    def diff(
        self,
        cib_old_file,
        cib_new_file,
        name="runner.cib.diff",
        stdout="resulting diff",
        stderr="",
        returncode=1,  # 0 -> old and new are the same, 1 -> old and new differ
    ):
        """
        Create a call for diffing two CIBs stored in two files
        string cib_old_file -- path to a file with an old CIB
        string cib_new_file -- path to a file with a new CIB
        string name -- key of the call
        string stdout -- resulting diff
        string stderr -- error returned from the diff process
        int returncode -- exit code of the diff process
        """
        self.__calls.place(
            name,
            RunnerCall(
                [
                    "crm_diff",
                    "--original",
                    cib_old_file,
                    "--new",
                    cib_new_file,
                    "--no-version",
                ],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def push_diff(
        self,
        name="runner.cib.push_diff",
        cib_diff="resulting diff",
        stdout="",
        stderr="",
        returncode=0,
        env=None,
    ):
        """
        Create a call for pushing a diff of CIBs
        string name -- key of the call
        string cib_diff -- the diff of CIBs
        dict env -- CommandRunner environment variables
        """
        self.__calls.place(
            name,
            RunnerCall(
                ["cibadmin", "--patch", "--verbose", "--xml-pipe"],
                check_stdin=CheckStdinEqualXml(cib_diff),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                env=env,
            ),
        )

    def upgrade(self, name="runner.cib.upgrade", before=None):
        """
        Create call for upgrading cib.

        string name -- key of the call
        string before -- key of call before which this new call is to be placed
        """
        self.__calls.place(
            name,
            RunnerCall(["cibadmin", "--upgrade", "--force"]),
            before=before,
        )



from collections import defaultdict
from dataclasses import replace as dt_replace
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
)

from lxml import etree
from lxml.etree import _Element

from pcs.common import (
    const,
    pacemaker,
    reports,
)
from pcs.common.pacemaker import role
from pcs.common.pacemaker.resource.operations import CibResourceOperationDto
from pcs.common.reports import (
    ReportItemList,
    ReportProcessor,
)
from pcs.common.reports.item import ReportItem
from pcs.common.tools import timeout_to_seconds
from pcs.common.types import StringCollection
from pcs.lib import validate
from pcs.lib.cib import (
    nvpair_multi,
    rule,
)
from pcs.lib.cib.nvpair import append_new_instance_attributes
from pcs.lib.cib.resource.agent import (
    complete_operations_options,
    get_default_operation_interval,
    operation_dto_to_legacy_dict,
)
from pcs.lib.cib.resource.const import OPERATION_ATTRIBUTES as ATTRIBUTES
from pcs.lib.cib.resource.types import (
    ResourceOperationFilteredIn,
    ResourceOperationFilteredOut,
    ResourceOperationIn,
)
from pcs.lib.cib.tools import (
    create_subelement_id,
    does_id_exist,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import is_true
from pcs.lib.tools import get_optional_value

OPERATION_NVPAIR_ATTRIBUTES = [
    "OCF_CHECK_LEVEL",
]

ON_FAIL_VALUES = tuple(
    sorted(
        map(
            str,
            (
                const.PCMK_ON_FAIL_ACTION_IGNORE,
                const.PCMK_ON_FAIL_ACTION_BLOCK,
                const.PCMK_ON_FAIL_ACTION_DEMOTE,
                const.PCMK_ON_FAIL_ACTION_STOP,
                const.PCMK_ON_FAIL_ACTION_RESTART,
                const.PCMK_ON_FAIL_ACTION_STANDBY,
                const.PCMK_ON_FAIL_ACTION_FENCE,
                const.PCMK_ON_FAIL_ACTION_RESTART_CONTAINER,
            ),
        )
    )
)

_BOOLEAN_VALUES = [
    "0",
    "1",
    "true",
    "false",
]

# _normalize(key, value) -> normalized_value
_normalize = validate.option_value_normalization(
    {
        "role": lambda value: value.capitalize(),
        "on-fail": lambda value: value.lower(),
        "record-pending": lambda value: value.lower(),
        "enabled": lambda value: value.lower(),
    }
)


def prepare(
    report_processor: ReportProcessor,
    raw_operation_list: Iterable[ResourceOperationFilteredIn],
    default_operation_list: Iterable[CibResourceOperationDto],
    allowed_operation_name_list: StringCollection,
    new_role_names_supported: bool,
    allow_invalid: bool = False,
) -> List[ResourceOperationFilteredOut]:
    """
    Return operation_list prepared from raw_operation_list and
    default_operation_list.

    report_processor -- tool for warning/info/error reporting
    raw_operation_list -- user entered operations that require follow-up care
    default_operation_list -- operations defined as default by (most probably)
        a resource agent
    allowed_operation_name_list -- operation names defined by a resource agent
    allow_invalid -- flag for validation skipping
    """
    operations_to_validate = _operations_to_normalized(raw_operation_list)

    report_list: ReportItemList = []
    report_list.extend(
        _validate_operation_list(
            operations_to_validate, allowed_operation_name_list, allow_invalid
        )
    )

    operation_list = _normalized_to_operations(
        operations_to_validate, new_role_names_supported
    )

    report_list.extend(validate_different_intervals(operation_list))

    if report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    report_list, remaining_default_operations = uniquify_operations_intervals(
        _get_remaining_defaults(
            operation_list,
            default_operation_list,
        )
    )
    if report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    return [
        _filter_op_dict(op, new_role_names_supported)
        for op in complete_operations_options(operation_list)
        + [
            operation_dto_to_legacy_dict(op, {})
            for op in remaining_default_operations
        ]
    ]


def _operations_to_normalized(
    raw_operation_list: Iterable[ResourceOperationFilteredIn],
) -> List[validate.TypeOptionNormalizedMap]:
    return [
        validate.values_to_pairs(op, _normalize) for op in raw_operation_list
    ]


def _normalized_to_operations(
    normalized_pairs: Iterable[validate.TypeOptionNormalizedMap],
    new_role_names_supported: bool,
) -> List[ResourceOperationFilteredOut]:
    def _replace_role(op_dict):
        if "role" in op_dict:
            op_dict["role"] = pacemaker.role.get_value_for_cib(
                op_dict["role"], new_role_names_supported
            )
        return op_dict

    return [
        _replace_role(validate.pairs_to_values(op)) for op in normalized_pairs
    ]


def _validate_operation_list(
    operation_list, allowed_operation_name_list, allow_invalid=False
):
    severity = reports.item.get_severity(reports.codes.FORCE, allow_invalid)
    option_type = "resource operation"

    validators = [
        validate.NamesIn(ATTRIBUTES, option_type=option_type),
        validate.IsRequiredAll(["name"], option_type=option_type),
        validate.ValueIn(
            "name",
            allowed_operation_name_list,
            option_name_for_report="operation name",
            severity=severity,
        ),
        validate.ValueIn("role", const.PCMK_ROLES),
        validate.ValueDeprecated(
            "role",
            {
                const.PCMK_ROLE_PROMOTED_LEGACY: const.PCMK_ROLE_PROMOTED,
                const.PCMK_ROLE_UNPROMOTED_LEGACY: const.PCMK_ROLE_UNPROMOTED,
            },
            reports.ReportItemSeverity.deprecation(),
        ),
        validate.ValueIn("on-fail", ON_FAIL_VALUES),
        validate.ValueIn("record-pending", _BOOLEAN_VALUES),
        validate.ValueIn("enabled", _BOOLEAN_VALUES),
        validate.MutuallyExclusive(
            ["interval-origin", "start-delay"], option_type=option_type
        ),
        validate.ValueId("id", option_name_for_report="operation id"),
    ]
    validator_all = validate.ValidatorAll(validators)

    report_list = []
    for operation in operation_list:
        report_list.extend(validator_all.validate(operation))
    return report_list


def _filter_op_dict(
    op: ResourceOperationIn, new_role_names_supported: bool
) -> ResourceOperationFilteredOut:
    # adjust new operation definition (coming from agent metadata) to old code,
    # TODO this should be handled in a different place - in saving operations to
    # CIB
    result = {
        key: val for key, val in op.items() if val is not None and val != ""
    }
    # translate a role to a proper value
    if "role" in result:
        result["role"] = pacemaker.role.get_value_for_cib(
            const.PcmkRoleType(result["role"]), new_role_names_supported
        )
    return result


def _get_remaining_defaults(
    operation_list: Iterable[ResourceOperationFilteredIn],
    default_operation_list: Iterable[CibResourceOperationDto],
) -> List[CibResourceOperationDto]:
    """
    Return operations not mentioned in operation_list but contained in
        default_operation_list.

    operation_list -- user entered operations that require follow-up care
    default_operation_list -- operations defined as default by (most probably)
        a resource agent
    """
    defined_operation_names = frozenset(
        operation["name"] for operation in operation_list
    )
    return [
        default_operation
        for default_operation in default_operation_list
        if default_operation.name not in defined_operation_names
    ]


def _get_interval_uniquer():
    used_intervals_map = defaultdict(set)

    def get_uniq_interval(name, initial_interval):
        """
        Return unique interval for name based on initial_interval if
        initial_interval is valid or return initial_interval otherwise.

        string name is the operation name for searching interval
        initial_interval is starting point for finding free value
        """
        used_intervals = used_intervals_map[name]
        normalized_interval = timeout_to_seconds(initial_interval)
        if normalized_interval is None:
            return initial_interval

        if normalized_interval not in used_intervals:
            used_intervals.add(normalized_interval)
            return initial_interval

        while normalized_interval in used_intervals:
            normalized_interval += 1
        used_intervals.add(normalized_interval)
        return str(normalized_interval)

    return get_uniq_interval


def op_element_to_dto(
    op_element: _Element, rule_eval: Optional[rule.RuleInEffectEval] = None
) -> CibResourceOperationDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    return CibResourceOperationDto(
        id=str(op_element.attrib["id"]),
        name=str(op_element.attrib["name"]),
        interval=str(op_element.attrib["interval"]),
        description=op_element.get("description"),
        start_delay=op_element.get("start-delay"),
        interval_origin=op_element.get("interval-origin"),
        timeout=op_element.get("timeout"),
        enabled=get_optional_value(is_true, op_element.get("enabled")),
        record_pending=get_optional_value(
            is_true, op_element.get("record-pending")
        ),
        role=get_optional_value(
            lambda _role: role.get_value_primary(const.PcmkRoleType(_role)),
            op_element.get("role"),
        ),
        on_fail=get_optional_value(
            const.PcmkOnFailAction, op_element.get("on-fail")
        ),
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                op_element, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                op_element, nvpair_multi.NVSET_INSTANCE
            )
        ],
    )


def uniquify_operations_intervals(
    operation_list: Iterable[CibResourceOperationDto],
) -> Tuple[reports.ReportItemList, List[CibResourceOperationDto]]:
    """
    Return list of operation where intervals for the same operation are unique

    operation_list -- operations the new operation list will be based on
    """
    get_unique_interval = _get_interval_uniquer()
    report_list = []
    new_operations = []
    for operation in operation_list:
        new_interval = get_unique_interval(operation.name, operation.interval)
        if new_interval != operation.interval:
            report_list.append(
                ReportItem.warning(
                    reports.messages.ResourceOperationIntervalAdapted(
                        operation.name,
                        operation.interval,
                        new_interval,
                    )
                )
            )
            operation = dt_replace(operation, interval=new_interval)
        new_operations.append(operation)
    return report_list, new_operations


def validate_different_intervals(operation_list):
    """
    Check that the same operations (e.g. monitor) have different interval.
    list operation_list contains dictionaries with attributes of operation
    return see resource operation in pcs/lib/exchange_formats.md
    """
    duplication_map = defaultdict(lambda: defaultdict(list))
    for operation in operation_list:
        interval = operation.get(
            "interval", get_default_operation_interval(operation["name"])
        )
        seconds = timeout_to_seconds(interval)
        duplication_map[operation["name"]][seconds].append(interval)

    duplications = defaultdict(list)
    for name, interval_map in duplication_map.items():
        for timeout in sorted(interval_map.values()):
            if len(timeout) > 1:
                duplications[name].append(timeout)

    if duplications:
        return [
            ReportItem.error(
                reports.messages.ResourceOperationIntervalDuplication(
                    dict(duplications)
                )
            )
        ]
    return []


def create_id(context_element, id_provider, name, interval):
    """
    Create id for op element.
    etree context_element is used for the name building
    string name is the name of the operation
    mixed interval is the interval attribute of operation
    """
    return create_subelement_id(
        context_element, "{0}-interval-{1}".format(name, interval), id_provider
    )


def create_operations(primitive_element, id_provider, operation_list):
    """
    Create operation element containing operations from operation_list

    list operation_list contains dictionaries with attributes of operation
    IdProvider id_provider -- elements' ids generator
    etree primitive_element is context element
    """
    operations_element = etree.SubElement(primitive_element, "operations")
    for operation in sorted(operation_list, key=lambda op: op["name"]):
        append_new_operation(operations_element, id_provider, operation)


def append_new_operation(operations_element, id_provider, options):
    """
    Create op element and append it to operations_element.

    etree operations_element is the context element
    IdProvider id_provider -- elements' ids generator
    dict options are attributes of operation
    """
    attribute_map = dict(
        (key, value)
        for key, value in options.items()
        if key not in OPERATION_NVPAIR_ATTRIBUTES
    )
    if "id" in attribute_map:
        if does_id_exist(operations_element, attribute_map["id"]):
            raise LibraryError(
                ReportItem.error(
                    reports.messages.IdAlreadyExists(attribute_map["id"])
                )
            )
    else:
        attribute_map.update(
            {
                "id": create_id(
                    operations_element.getparent(),
                    id_provider,
                    options["name"],
                    options["interval"],
                )
            }
        )
    op_element = etree.SubElement(
        operations_element,
        "op",
        attribute_map,
    )
    nvpair_attribute_map = dict(
        (key, value)
        for key, value in options.items()
        if key in OPERATION_NVPAIR_ATTRIBUTES
    )

    if nvpair_attribute_map:
        append_new_instance_attributes(
            op_element, nvpair_attribute_map, id_provider
        )

    return op_element


def get_resource_operations(resource_el, names=None):
    """
    Get operations of a given resource, optionally filtered by name
    etree resource_el -- resource element
    iterable names -- return only operations of these names if specified
    """
    return [
        op_el
        for op_el in resource_el.xpath("./operations/op")
        if not names or op_el.attrib.get("name", "") in names
    ]


def disable(operation_element):
    """
    Disable the specified operation
    etree operation_element -- the operation
    """
    operation_element.attrib["enabled"] = "false"


def enable(operation_element):
    """
    Enable the specified operation
    etree operation_element -- the operation
    """
    operation_element.attrib.pop("enabled", None)


def is_enabled(operation_element):
    """
    Check if the specified operation is enabled
    etree operation_element -- the operation
    """
    return is_true(operation_element.attrib.get("enabled", "true"))



import os

from pcs import settings
from pcs.common import pcs_pycurl as pycurl
from pcs.common import reports
from pcs.common.node_communicator import (
    CommunicatorLoggerInterface,
    HostNotFound,
    NodeTargetFactory,
)
from pcs.common.reports import (
    ReportItemSeverity,
    ReportProcessor,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.errors import LibraryError


def _get_port(port):
    return port if port is not None else settings.pcsd_default_port


class LibCommunicatorLogger(CommunicatorLoggerInterface):
    def __init__(self, logger, reporter: ReportProcessor):
        self._logger = logger
        self._reporter = reporter

    def log_request_start(self, request):
        msg = "Sending HTTP Request to: {url}"
        if request.data:
            msg += "\n--Debug Input Start--\n{data}\n--Debug Input End--"
        self._logger.debug(msg.format(url=request.url, data=request.data))
        self._reporter.report(
            ReportItem.debug(
                reports.messages.NodeCommunicationStarted(
                    request.url,
                    request.data,
                )
            )
        )

    def log_response(self, response):
        if response.was_connected:
            self._log_response_successful(response)
        else:
            self._log_response_failure(response)
        self._log_debug(response)

    def _log_response_successful(self, response):
        url = response.request.url
        msg = (
            "Finished calling: {url}\nResponse Code: {code}"
            + "\n--Debug Response Start--\n{response}\n--Debug Response End--"
        )
        self._logger.debug(
            msg.format(
                url=url, code=response.response_code, response=response.data
            )
        )
        self._reporter.report(
            ReportItem.debug(
                reports.messages.NodeCommunicationFinished(
                    url,
                    response.response_code,
                    response.data,
                )
            )
        )

    def _log_response_failure(self, response):
        msg = "Unable to connect to {node} ({reason})"
        self._logger.debug(
            msg.format(
                node=response.request.host_label, reason=response.error_msg
            )
        )
        self._reporter.report(
            ReportItem.debug(
                reports.messages.NodeCommunicationNotConnected(
                    response.request.host_label,
                    response.error_msg,
                )
            )
        )
        if is_proxy_set(os.environ):
            self._logger.warning("Proxy is set")
            self._reporter.report(
                ReportItem.warning(
                    reports.messages.NodeCommunicationProxyIsSet(
                        response.request.host_label,
                        response.request.dest.addr,
                    )
                )
            )

    def _log_debug(self, response):
        url = response.request.url
        debug_data = response.debug
        self._logger.debug(
            (
                "Communication debug info for calling: {url}\n"
                "--Debug Communication Info Start--\n"
                "{data}\n"
                "--Debug Communication Info End--"
            ).format(url=url, data=debug_data)
        )
        self._reporter.report(
            ReportItem.debug(
                reports.messages.NodeCommunicationDebugInfo(url, debug_data)
            )
        )

    def log_retry(self, response, previous_dest):
        old_port = _get_port(previous_dest.port)
        new_port = _get_port(response.request.dest.port)
        msg = (
            "Unable to connect to '{label}' via address '{old_addr}' and port "
            "'{old_port}'. Retrying request '{req}' via address '{new_addr}' "
            "and port '{new_port}'"
        ).format(
            label=response.request.host_label,
            old_addr=previous_dest.addr,
            old_port=old_port,
            new_addr=response.request.dest.addr,
            new_port=new_port,
            req=response.request.url,
        )
        self._logger.warning(msg)
        self._reporter.report(
            ReportItem.warning(
                reports.messages.NodeCommunicationRetrying(
                    response.request.host_label,
                    previous_dest.addr,
                    old_port,
                    response.request.dest.addr,
                    new_port,
                    response.request.url,
                )
            )
        )

    def log_no_more_addresses(self, response):
        msg = "No more addresses for node {label} to run '{req}'".format(
            label=response.request.host_label,
            req=response.request.url,
        )
        self._logger.warning(msg)
        self._reporter.report(
            ReportItem.warning(
                reports.messages.NodeCommunicationNoMoreAddresses(
                    response.request.host_label,
                    response.request.url,
                )
            )
        )


class NodeTargetLibFactory(NodeTargetFactory):
    def __init__(self, known_hosts, report_processor: ReportProcessor):
        super().__init__(known_hosts)
        self._report_processor = report_processor

    def get_target_list_with_reports(
        self,
        host_name_list,
        skip_non_existing=False,
        allow_skip=True,
        report_none_host_found=True,
    ):
        target_list = []
        unknown_host_list = []
        for host_name in host_name_list:
            try:
                target_list.append(self.get_target(host_name))
            except HostNotFound:
                unknown_host_list.append(host_name)

        report_list = []
        if unknown_host_list:
            report_list.append(
                ReportItem(
                    severity=reports.item.get_severity(
                        (
                            reports.codes.SKIP_OFFLINE_NODES
                            if allow_skip
                            else None
                        ),
                        skip_non_existing,
                    ),
                    message=reports.messages.HostNotFound(
                        sorted(unknown_host_list),
                    ),
                )
            )

        if not target_list and host_name_list and report_none_host_found:
            # we want to create this report only if there was at least one
            # required address specified
            report_list.append(
                ReportItem.error(reports.messages.NoneHostFound())
            )
        return report_list, target_list

    def get_target_list(
        self, host_name_list, skip_non_existing=False, allow_skip=True
    ):
        report_list, target_list = self.get_target_list_with_reports(
            host_name_list, skip_non_existing, allow_skip
        )
        if report_list:
            if self._report_processor.report_list(report_list).has_errors:
                raise LibraryError()
        return target_list


def response_to_report_item(
    response,
    severity=ReportItemSeverity.ERROR,
    forceable=None,
    report_pcsd_too_old_on_404=False,
):
    """
    Returns report item which corresponds to response if was not successful.
    Otherwise returns None.

    Response response -- response from which report item shoculd be created
    ReportItemseverity severity -- severity of report item
    string forceable -- force code
    bool report_pcsd_too_old_on_404 -- if False, report unsupported command
    """
    response_code = response.response_code
    report_item = None
    reason = None
    if (
        report_pcsd_too_old_on_404
        and response.was_connected
        and response_code == 404
    ):
        return ReportItem.error(
            reports.messages.PcsdVersionTooOld(response.request.host_label)
        )
    if response.was_connected:
        if response_code == 400:
            # old pcsd protocol: error messages are commonly passed in plain
            # text in response body with HTTP code 400
            # we need to be backward compatible with that
            report_item = reports.messages.NodeCommunicationCommandUnsuccessful
            reason = response.data.rstrip()
        elif response_code == 401:
            report_item = reports.messages.NodeCommunicationErrorNotAuthorized
            reason = "HTTP error: {0}".format(response_code)
        elif response_code == 403:
            report_item = (
                reports.messages.NodeCommunicationErrorPermissionDenied
            )
            reason = "HTTP error: {0}".format(response_code)
        elif response_code == 404:
            report_item = (
                reports.messages.NodeCommunicationErrorUnsupportedCommand
            )
            reason = "HTTP error: {0}".format(response_code)
        elif response_code >= 400:
            report_item = reports.messages.NodeCommunicationError
            reason = "HTTP error: {0}".format(response_code)
    else:
        if response.errno in [
            pycurl.E_OPERATION_TIMEDOUT,
            pycurl.E_OPERATION_TIMEOUTED,
        ]:
            report_item = reports.messages.NodeCommunicationErrorTimedOut
            reason = response.error_msg
        else:
            report_item = reports.messages.NodeCommunicationErrorUnableToConnect
            reason = response.error_msg
    if not report_item:
        return None
    return ReportItem(
        severity=ReportItemSeverity(severity, forceable),
        message=report_item(
            response.request.host_label,
            response.request.action,
            reason,
        ),
    )


def is_proxy_set(env_dict):
    """
    Returns True whenever any of proxy environment variables (https_proxy,
    HTTPS_PROXY, all_proxy, ALL_PROXY) are set in env_dict. False otherwise.

    dict env_dict -- environment variables in dict
    """
    proxy_list = ["https_proxy", "all_proxy"]
    for var in proxy_list + [v.upper() for v in proxy_list]:
        if env_dict.get(var, "") != "":
            return True
    return False
